'use strict';

var types = require('babel-types');
var template = require('babel-template');
var urlJoin = require('../utils/urlJoin');
var isURL = require('../utils/is-url');
var matchesPattern = require('./matches-pattern');

var requireTemplate = template('require("_bundle_loader")');
var argTemplate = template('require.resolve(MODULE)');
var serviceWorkerPattern = ['navigator', 'serviceWorker', 'register'];

module.exports = {
  ImportDeclaration: function ImportDeclaration(node, asset) {
    asset.isES6Module = true;
    addDependency(asset, node.source);
  },
  ExportNamedDeclaration: function ExportNamedDeclaration(node, asset) {
    asset.isES6Module = true;
    if (node.source) {
      addDependency(asset, node.source);
    }
  },
  ExportAllDeclaration: function ExportAllDeclaration(node, asset) {
    asset.isES6Module = true;
    addDependency(asset, node.source);
  },
  ExportDefaultDeclaration: function ExportDefaultDeclaration(node, asset) {
    asset.isES6Module = true;
  },
  CallExpression: function CallExpression(node, asset) {
    var callee = node.callee,
        args = node.arguments;


    var isRequire = types.isIdentifier(callee) && callee.name === 'require' && args.length === 1 && types.isStringLiteral(args[0]);

    if (isRequire) {
      addDependency(asset, args[0]);
      return;
    }

    var isDynamicImport = callee.type === 'Import' && args.length === 1 && types.isStringLiteral(args[0]);

    if (isDynamicImport) {
      asset.addDependency('_bundle_loader');
      addDependency(asset, args[0], { dynamic: true });

      node.callee = requireTemplate().expression;
      node.arguments[0] = argTemplate({ MODULE: args[0] }).expression;
      asset.isAstDirty = true;
      return;
    }

    var isRegisterServiceWorker = types.isStringLiteral(args[0]) && matchesPattern(callee, serviceWorkerPattern);

    if (isRegisterServiceWorker) {
      addURLDependency(asset, args[0]);
      return;
    }
  },
  NewExpression: function NewExpression(node, asset) {
    var callee = node.callee,
        args = node.arguments;


    var isWebWorker = callee.type === 'Identifier' && callee.name === 'Worker' && args.length === 1 && types.isStringLiteral(args[0]);

    if (isWebWorker) {
      addURLDependency(asset, args[0]);
      return;
    }
  }
};

function addDependency(asset, node) {
  var opts = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  opts.loc = node.loc && node.loc.start;
  asset.addDependency(node.value, opts);
}

function addURLDependency(asset, node) {
  var assetPath = asset.addURLDependency(node.value);
  if (!isURL(assetPath)) {
    assetPath = urlJoin(asset.options.publicURL, assetPath);
  }
  node.value = assetPath;
  asset.isAstDirty = true;
}